/*
 * kgraft_patch_bsc1185899
 *
 * Fix for CVE-2021-32399, bsc#1185899
 *
 *  Upstream commit:
 *  e2cb6b891ad2 ("bluetooth: eliminate the potential race condition when
 *                 removing the HCI controller")
 *
 *  SLE12-SP3 commit:
 *  12d067d47245adf44b7ed85f4d752157fe0818af
 *
 *  SLE12-SP4, SLE15 and SLE15-SP1 commit:
 *  4b51cab85bfc062d1abace713e95252760b8ca29
 *
 *  SLE12-SP5 commit:
 *  089b28d63c3c114bef838b99d201590eeaab5b73
 *
 *  SLE15-SP2 commit:
 *  3c88de9e7a9b84d6124e3b97abd13d6ee54bbdf3
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if !IS_MODULE(CONFIG_BT)
#error "Live patch supports only CONFIG_BT=m"
#endif

/* klp-ccp: from net/bluetooth/hci_core.c */
#include <linux/export.h>
#include <linux/idr.h>
#include <linux/rfkill.h>
#include <linux/debugfs.h>
#include <linux/crypto.h>
#include <net/bluetooth/bluetooth.h>
#include <net/bluetooth/hci_core.h>

/* klp-ccp: from net/bluetooth/hci_request.h */
struct hci_request;

/* klp-ccp: from net/bluetooth/hci_core.c */
#define hci_req_lock(d)		mutex_lock(&d->req_lock)
#define hci_req_unlock(d)	mutex_unlock(&d->req_lock)

static int (*kgre___hci_req_sync)(struct hci_dev *hdev,
			  void (*func)(struct hci_request *req,
				      unsigned long opt),
			  unsigned long opt, __u32 timeout);

int kgrp_hci_req_sync(struct hci_dev *hdev,
			void (*req)(struct hci_request *req,
				    unsigned long opt),
			unsigned long opt, __u32 timeout)
{
	int ret;

	/*
	 * Fix CVE-2021-32399
	 *  -2 lines
	 */

	/* Serialize all requests */
	hci_req_lock(hdev);
	/*
	 * Fix CVE-2021-32399
	 *  -1 line, +8 lines
	 */
	/* check the state after obtaing the lock to protect the HCI_UP
	 * against any races from hci_dev_do_close when the controller
	 * gets removed.
	 */
	if (test_bit(HCI_UP, &hdev->flags))
		ret = (*kgre___hci_req_sync)(hdev, req, opt, timeout);
	else
		ret = -ENETDOWN;
	hci_req_unlock(hdev);

	return ret;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1185899.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "bluetooth"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "__hci_req_sync", (void *)&kgre___hci_req_sync, "bluetooth" },
};

static int kgr_patch_bsc1185899_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1185899_module_nb = {
	.notifier_call = kgr_patch_bsc1185899_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1185899_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1185899_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1185899_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1185899_module_nb);
}
